# -*- Python -*-
# $Id: actions.py,v 1.15 2004/02/15 20:22:05 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, October 2001
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""Actions to take, for use in config file.

Most importantly, I define L{CommitAction}, which you need to subclass if
you want to be an action in a
L{ConfigurationSet<cvstoys.config.ConfigurationSet>}.

Also, any class in this module defined as a subclass of L{CommitAction}
will find itself included in the global namespace in which
configuration files are executed.
"""

__version__ = '$Revision: 1.15 $'[11:-2]

# TODO:
#  - plug-in system for actions?
#  - Action isa ApplicationService, configuration is a MultiService.

# Actions for:
#  - pychecker
#  - unit tests
#  - Human-readable Python-tailored diffs for the notification mails.
#    Instead of saying "the seven lines around line 313", it will
#    identify locations in the source with "Foo.__init__: [...]".

from twisted.python.compat import *

from twisted.internet import reactor
from twisted.internet import app, defer, error, protocol
from twisted.python import components, failure, log

import errno, os, string
from os import path

from cvstoys import interfaces

class CommitAction:
    """(Abstract)"""
    __implements__ = (interfaces.ICommitAction,)

    def __call__(self, root, files, message, user):
        return defer.succeed(None)

    def unload(self):
        return defer.succeed(None)

class DebugAction(CommitAction):
    """I dump crap to the log."""

    def __init__(self, foo=''):
        """You may supply a foo to appear in each message dumped."""
        self.foo = foo

    def __call__(self, root, files, message, user):
        x = 'x' * 30
        s = ("%s %s %s %s\n"
             "%s\n"
             "%s MESSAGE %s\n%s"
             % (x, self.foo, root, user,
                ' '.join(map(str, files)),
                x,x, message or "(no message)\n"))
        log.msg(s)

        return CommitAction.__call__(self, root, files, message, user)


class Serial(CommitAction):
    """A list of actions which take place in serial.
    """
    # Actually a factory which creates and runs a SerialString when called.

    def __init__(self, actionList):
        """Made of a list of actions.
        """
        self.actionList = actionList

    def __call__(self, *args, **kw):
        s = SerialString(self.actionList)
        apply(s, args, kw)
        return s

    def __len__(self):
        return len(self.actionList)


class ExecFile(CommitAction):
    def __init__(self, filename):
        self.filename = filename

    def __call__(self, *unused, **unused_kw):
        d = defer.Deferred()
        try:
            execfile(self.filename, {})
        except:
            d.errback(failure.Failure())
        else:
            d.callback(None)
        return d


class CheckOut(CommitAction):
    """I will keep a working copy of the repository up to date.
    """

    def __init__(self, working_root):
        """Check it out.

        I take one argument -- the directory I will be placing checked
        out modules in.
        """

        if not path.isabs(working_root):
            log.msg("warning: working directory '%s' not an absolute path"
                    % working_root)

        if not path.exists(working_root):
            log.msg("working directory '%s' does not exist; creating it."
                    % working_root)
            os.makedirs(working_root)

        if not path.isdir(working_root):
            raise OSError(errno.ENOTDIR, os.strerror(errno.ENOTDIR),
                          working_root)

        if not os.access(working_root, os.X_OK):
            raise OSError(errno.EACCES, os.strerror(errno.EACCES),
                          working_root)

        if not os.access(working_root, os.W_OK):
            log.msg("No write permission to working directory '%s'.\n"
                    "I might be able to update existing working copies of"
                    " modules, but not check out new ones."
                    % working_root)

        self.workRoot = working_root


    def __call__(self, root, files, unused_message, unused_user):
        by_module = {}

        # sort files by module
        for fil in files:
            moduleName = fil.pathname.split('/', 1)[0]

            if not by_module.has_key(moduleName):
                by_module[moduleName] = []
            by_module[moduleName].append(fil)

        deferreds = []
        for (moduleName, mfiles) in by_module.items():
            # Has this module already been checked out?
            CVS = path.join(self.workRoot, moduleName, 'CVS')
            if not (path.exists(CVS) and
                    path.exists(path.join(CVS,'Repository')) and
                    path.exists(path.join(CVS,'Root'))):

                deferreds.append(cvsCommand('checkout', root,
                                            self.workRoot, moduleName))
            else:
                if not os.access(CVS, os.W_OK | os.X_OK):
                    raise OSError(errno.EACCES, os.strerror(errno.EACCES),
                                  CVS)

                paths = [f.pathname for f in mfiles]
                common = commonprefix(paths)
                # NOTE: We just do cvs update from the common path and
                #   let it do its recursive thing.
                # As an alternative, we could walk the directories and
                # update files explicitly, but...  why bother?

                workdir = path.join(self.workRoot, common)
                # It's possible this directory does not exist, if for instance
                # these are the first files added to the directory.
                workdir = existingBase(workdir)
                deferreds.append(cvsCommand('update', root, workdir))
        return defer.DeferredList(deferreds)


class Rsync(CommitAction):
    """Mirror the CVS repository with rsync.

    This mirrors the server repository, not a working copy.
    """
    def __init__(self, destination):
        """Constructor for rsyncing to a remote destination.

        'destination' is a string user@host:path, such as you would use
        as a destination path with rsync.
        """
        self.destination = destination

    def __call__(self, root, files, unused_message, unused_user):
        "rsync -a module %(dest)s"
        by_module = {}

        # sort files by module
        for fil in files:
            moduleName = fil.pathname.split('/', 1)[0]
            # moduleName = string.split(fil.pathname)[0]

            if not by_module.has_key(moduleName):
                by_module[moduleName] = []
            by_module[moduleName].append(fil)

        deferreds = []
        for (moduleName, mfiles) in by_module.items():
            paths = map(lambda f: f.pathname, mfiles)
            common = commonprefix(paths)
            deferreds.append(rsyncCommand(root, common, self.destination))

        return defer.DeferredList(deferreds)


######################################################################

class SerialString(defer.Deferred):
    """XXX - class naming needs work!
    """
    def __init__(self, actionList):
        """Made of a list of actions.
        """
        defer.Deferred.__init__(self)
        self.offset = 0
        self.actionList = actionList

    def __call__(self, *args, **kw):
        self.args = args
        self.kw = kw
        self.next()

    def next(self, result=None):
        if self.offset >= len(self):
            # Release the list reference now that we're done.
            self.actionList = ()
            self.callback(result)
        else:
            thisAction = self.actionList[self.offset]
            self.offset += 1
            d = apply(thisAction, self.args, self.kw)
            d.addCallbacks(self.next, self.errback)

    def __len__(self):
        return len(self.actionList)


class BasicProcessProtocol(protocol.ProcessProtocol):
    def __init__(self):
        # super(BasicProcessProtocol, self).__init__(*a, **kw)
        self.deferred = defer.Deferred()

    def outReceived(self, data):
        # XXX
        log.msg(data)
    def errReceived(self, data):
        # XXX
        log.msg(data)

    def processEnded(self, reason):
        # XXX: If I have done errReceived, maybe I should errback instead
        # of callback.
        if reason.check(error.ProcessDone):
            self.deferred.callback(None)
        else:
            self.deferred.errback(reason)


def cvsCommand(cvs_command, cvsroot, curdir,
               moduleName=None, readonly=True, noisy=False):
    """I run a CVS command.

    Commands I support are 'checkout' and 'update'.

    The moduleName is unused for 'update' -- I just run update in
    whatever directory it is that I'm told to be in.

    I return a Deferred, so you can know when the command is done.
    """

    args = ['cvs']
    if readonly:
        args.append('-r')
    if not noisy:
        args.append('-Q')

    if cvs_command in ('checkout','co','get'):
        if not moduleName:
            raise TypeError,\
                  "moduleName must be supplied for checkout"
        args.extend(['checkout', moduleName])

    elif cvs_command in ('update','up','upd'):
        args.extend(['update', '-dP'])

    else:
        raise ValueError, "Unsupported CVS command", cvs_command

    environment = os.environ.copy()
    environment["CVSROOT"] = cvsroot

    proto = BasicProcessProtocol()
    reactor.spawnProcess(proto, "cvs", args, environment, curdir)
    return proto.deferred


def rsyncCommand(path, source, destination):
    args = ["rsync", "--archive", "--quiet", "--relative",
            "--exclude", "#cvs.*",
            source, destination]
    environment = os.environ.copy()
    proto = BasicProcessProtocol()
    reactor.spawnProcess(proto, "rsync", args, environment, path)
    return proto.deferred


# Actions in sibling modules.
# ARGH can't do this here.  It works fine if actions.actions is imported first,
# then when actions.pb imports actions it gets the already loaded module.
# But when actions.pb is imported first, it imports this module before it
# defines PBService, and the import gets that not-fully-made module...  grr.
#import pb
#from pb import PBService
#from mail import MailNotification
# ...so instead, we're doing "import actions; actions.Foo = Foo"
# in pb and mail.

#### Path utilities.

def commonprefix(paths):
    """non-broken L{os.path.commonprefix}.

    >>> paths = ['foo/bar/duck', 'foo/baz/goose']
    >>> os.path.commonprefix(paths)
    'foo/ba'
    >>> commonprefix(paths)
    'foo'

    """
    # Yes, this uses split('/') instead of disassembling the string by repeated
    # invocations of path.split, which might be the most correct way to do it.
    # I don't know anyone who runs CVS on a non-posix platform.
    paths = [p.split('/') for p in paths]
    # XXX: Apparently this "pass a list of lists to paths.commonprefix" is
    # undocumented behaviour, and there's some concern that it's prone to go
    # away because of that.  Send in a documentation patch to python so that
    # they'll feel bad about breaking this behavior!
    common = '/'.join(path.commonprefix(paths))
    return common

def existingBase(apath):
    while not path.exists(apath):
        apath=path.split(apath)[0]
        if not apath:
            raise ValueError, "Path whittled away to nothing."
    return apath


# Local Variables:
# elide-head-headers-to-hide:(("<license" . "</license>"))
# End:
