# -*- Python -*-
# $Id: mail.py,v 1.15 2003/04/19 21:11:21 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, October 2001
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""Send notification by e-mail, with diffs.

@var diffOpts: Switches to use with cvs diff.
@type diffOpts: list of strings
@var SENDMAIL: The pathname for the sendmail program.
"""

__version__ = '$Revision: 1.15 $'[11:-2]

SENDMAIL="/usr/lib/sendmail"

import actions

from cvstoys import common, util

from twisted.internet import defer, error, reactor

import os, string
from os import path

try:
    True
except NameError:
    True = (1==1)
    False = not True

# TODO:
#   Make mail-sending a bit more robust, instead of spawning a 'mail'
# process and ignoring the output and exit code.

# Questionable Ideas:
#   Run this from an action?  Then mail wouldn't have to originate
# from the CVS host.  But the mail host would have to get the diffs
# from CVS over the network.  Perhaps not a worthwhile, as we don't
# know anyone who actually needs to do that.

class MailNotification(actions.CommitAction):
    """I will send commit notifications by e-mail.

    I sing and dance and make diffs and URLs to ViewCVS.

    Please note that I am NOT a email list manager, nor do I intend to
    be.  While I will send to multiple recipients, I suggest that you
    just set me to post to a mailing list, and thus maintain the list
    of recipients with your list server's interface.
    """

    # viewcvs = "http://twistedmatrix.com/users/carmstro.twistd/static/viewcvs/cgi/viewcvs.cgi/"
    viewcvs = None
    replyTo = None
    staticFrom = None

    def __init__(self, recipientList=None, viewcvs=None, replyTo=None,
                 staticFrom=None):
        """Send mail to these guys.

        @param recipientList: e-mail addresses to send to.
        @type recipientList: List of Strings
        @param viewcvs: base url of a ViewCVS web service to make references to.
        @type viewcvs: string
        @param replyTo: Add a Reply-To address to the message headers.
        @type replyTo: string
        @param staticFrom: The From-address of all e-mails.  If omitted or
        C{None}, use the username of the cvs user making the commit.
        @type staticFrom: string
        """

        if recipientList is None:
            recipientList = []

        recipients = {}
        for r in recipientList:
            recipients[r] = 'Set'

        self.recipients = recipients

        if viewcvs:
            self.viewcvs = viewcvs
        if replyTo:
            self.replyTo = replyTo
        if staticFrom:
            self.staticFrom = staticFrom

    def addRecipient(self, address):
        self.recipients[address] = 'Set'

    def delRecipient(self, address):
        del self.recipients[address]

    def listRecipients(self):
        return self.recipients.keys()

    def __call__(self, root, files, message, user):
        return MailNotificationRequest(self, root, files, message, user)


# Install in 'actions' namespace so it gets picked up by configatron.
actions.MailNotification = MailNotification

class MailNotificationRequest(defer.Deferred):
    """(internal) When I grow up, I'll be a MailNotification message.

    I hold the state a MailNotification needs between when it receives
    its __call__ and when it can sendMessage.
    """

    urls = None

    def __init__(self, notifier, root, files, message, user):
        defer.Deferred.__init__(self)
        self.notifier = notifier
        self.cvsroot = root
        self.files = files
        self.logMessage = message
        self.user = user

        if message:
            # find the first non-blank line
            self.subject = message.lstrip().split('\n', 1)[0]
        else:
            self.subject = "commit in CVS %s" % root

        if notifier.viewcvs:
            self.urls = string.join(util.viewcvs_urls(notifier.viewcvs,
                                                      root, files), '\n')

        self.diffCollection = []

        d = self.makeDiffs()
        d.addCallback(self.sendMessage)

    def makeDiffs(self):
        deferreds = []
        for f in self.files:
            if isinstance(f, common.Directory):
                continue
            fullFilePath = path.join(f.pathname, f.filename)

            dontDiff = False
            revs = [f.r1, f.r2]
            for i in (0,1):
                if revs[i] in (None, (None,)):
                    # XXX: Need to clean up handling of unversioned things.
                    dontDiff = True
                else:
                    revs[i] = string.join(map(str, revs[i]), '.')

            if not dontDiff:
                d = cvsDiff(self.cvsroot, None, fullFilePath,
                            revs[0], revs[1])
                d.addCallback(self.addDiff)
                deferreds.append(d)

        # XXX: Is this a bug in DeferredList?
        if not deferreds:
            return defer.succeed(None)
        else:
            return defer.DeferredList(deferreds)

    def addDiff(self, diff):
        """call me back with a diff, I'll add it to the message.
        """
        self.diffCollection.append(diff)

    def sendMessage(self, unused=None):
        """Call me when all pieces of the message are collected.
        """

        msgChunks = ["Modified files:"]

        for f in self.files:
            if isinstance(f, common.VersionedPatch):
                msgChunks.append("%s %s %s"
                                 % (path.join(f.pathname, f.filename),
                                    string.join(map(str, f.r1 or (0,)),'.'),
                                    string.join(map(str, f.r2 or (0,)),'.')))
            else:
                msgChunks.append(path.join(f.pathname, f.filename))

        msgChunks.extend(['',
                          'Log message:',
                          self.logMessage])

        if self.urls:
            msgChunks.extend(['ViewCVS links:',
                              self.urls,
                              ''])

        msgChunks.extend(self.diffCollection)

        extra_headers = []
        if self.notifier.staticFrom is None:
            extra_headers.append('From: %s CVS <%s>' % (self.user, self.user,))
        else:
            extra_headers.append('From: %s' % (self.notifier.staticFrom,))
            msgChunks.insert(0,"Commit by: %s" % (self.user,))

        if self.notifier.replyTo:
            extra_headers.append('Reply-To: %s' % (self.notifier.replyTo,))

        msg = string.join(msgChunks, '\n')
        d = mailCommand(self.notifier.listRecipients(),
                        self.subject, msg, extra_headers)

        d.chainDeferred(self)

        del self.diffCollection

diffOpts = ['-u']
def cvsDiff(cvsroot, curdir, fil, v1, v2=None, noisy=False):
    args = ['cvs']

    if not noisy:
        args.append('-Q')

    args.extend(['-d',cvsroot])
    args.append('rdiff')

    args.extend(['-r',v1])
    if v2:
        args.extend(['-r', v2])

    args.extend(diffOpts)
    args.append(fil)

    environment = os.environ.copy()
    environment["CVSROOT"] = cvsroot

    proto = CvsDiffProtocol()
    reactor.spawnProcess(proto, "cvs", args, environment, curdir)
    return proto.deferred


class CvsDiffProtocol(actions.BasicProcessProtocol):
    def __init__(self):
        actions.BasicProcessProtocol.__init__(self)
        self.__data_chunks = []

    def outReceived(self, output):
        self.__data_chunks.append(output)

    def processEnded(self, reason):
        # XXX: If I have done errReceived, maybe I should errback instead
        # of callback.
        if reason.check(error.ProcessDone):
            s = string.join(self.__data_chunks, '')
            del self.__data_chunks
            self.deferred.callback(s)
        else:
            self.deferred.errback(reason)


def mailCommand(recipients, subject, content, extra_headers=None):
    args = ['sendmail', '-bm', '-t']
    headerList = []
    for r in recipients:
        headerList.append('To: %s' % r)
    if extra_headers:
        headerList.extend(extra_headers)
    # TODO: Include version in X-Mailer header.
    headerList.append("X-Mailer: CVSToys")
    headerList.append("Subject: %s" % (subject,))

    msg = '%s\n\n%s\n' % ('\n'.join(headerList), content)

    environment = os.environ.copy()
    environment["LONGNAME"] = "CVS"

    proto = actions.BasicProcessProtocol()
    p = reactor.spawnProcess(proto, SENDMAIL, args, environment, None)
    p.write(msg)
    p.loseConnection()
    return proto.deferred


### Local Variables:
### elide-head-headers-to-hide:(("<license" . "</license>"))
### End:
