# -*- Python -*-
# $Id: pb.py,v 1.28 2004/03/05 07:27:46 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, August 2002
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""A Perspective Broker service to publish notifications."""

__version__ = '$Revision: 1.28 $'[11:-2]

from twisted.cred import checkers, error, portal
from twisted.application import strports
from twisted.application.service import IService, IServiceCollection, \
     MultiService
from twisted.internet import app, defer, reactor
from twisted.persisted import styles
from twisted.python import components, log
from twisted.spread import pb

from cvstoys import interfaces
from cvstoys.clients.pb import PerspectiveInUse

# siblings
import actions

# system
import re

_matchAll = re.compile('')

class Notifiee(pb.Avatar, styles.Versioned):
    persistenceVersion = 2

    filter = None
    client = None

    def setFilter(self, root=None, path_reString=None, file_reString=None):
        """Set a filter for what notifications to receive.

        XXX: UNTESTED
        C{None} may be passed to any of the parameters to mean \"match all\".

        @param root: The name
        @type root: String
        @param path_reString: A regluar expression to match the path.
        @type path_reString: String
        @param file_reString: A regular expression to match the file name.
        @type file_reString: String
        """
        self.filter['root'] = root
        reStrings = {'path_re': path_reString,
                     'file_re': file_reString}

        for (name, rString) in reStrings.items():
            if rString is None:
                self.filter[name] = _matchAll
            else:
                self.filter[name] = re.compile(rString)

    perspective_setFilter = setFilter

    def notify(self, root, files, message, user):
        """Notify my client of a commit.

        Subject to the filter put in place by L{setFilter}.

        Arguments follow those of
        L{CommitAction.__call__<interfaces.ICommitAction.__call__>}
        """
        if self.filter:
            if self.filter['root'] and (root != self.filter['root']):
                return
            filteredfiles = []
            for f in files:
                if self.filter['path_re'].search(f.pathname) is None:
                    continue
                if self.filter['file_re'].search(f.filename) is None:
                    continue
                filteredfiles.append(f)
            if not filteredfiles:
                return
            else:
                files = filteredfiles
                del filteredfiles

        self.client[0].callRemote("notify", root, files, message, user)

    def goodbye(self, message):
        self.client[0].callRemote("goodbye", message)

    def attached(self, ref, identity):
        self.client = (ref, identity, None)
        return self

    def detached(self, unused_ref, unused_identity):
        del self.client

    def __getstate__(self):
        state = styles.Versioned.__getstate__(self)
        try:
            del state['client']
        except KeyError:
            pass
        return state


class _PBRealm(components.Adapter):
    __implements__ = portal.IRealm,
    perspectiveClass = Notifiee

    def __init__(self, original):
        components.Adapter.__init__(self, original)
        self.notifiees = {}

    def requestAvatar(self, avatarId, mind, *interfaces):
        av = self.perspectiveClass()
        av.attached(mind, avatarId)
        self.notifiees[av] = 'Set'
        def detached():
            av.detached(mind, avatarId)
            del self.notifiees[av]
        return (pb.IPerspective, av, detached)

    def connectedClients(self):
        return self.notifiees.keys()

    def sendGoodbyes(self, message):
        for n in self.notifiees.keys():
            n.goodbye(message)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['notifiees']
        return state

    def __setstate__(self, state):
        if 'notifiees' not in state.keys():
            self.notifiees = {}


class _PBService(MultiService, components.Adapter, styles.Versioned):
    persistenceVersion = 2

    def __init__(self, original):
        components.Adapter.__init__(self, original)
        MultiService.__init__(self)

    def stopService(self):
        portal.IRealm(self).sendGoodbyes("This PBService shutting down.")
        return MultiService.stopService(self)

    def upgradeToVersion2(self):
        # FIXME
        raise NotImplementedError

    def __getstate__(self):
        # FIXME: Boy oh boy do I want to use super() here.
        state = MultiService.__getstate__(self)
        state = styles.Versioned.__getstate__(self, state)
        return state


class PBService(components.Componentized, actions.CommitAction):
    """PB service to provide notification to remote agents.
    """
    portal = None
    netservice = None

    def __init__(self, serviceName="cvstoys.notify", userpass=None,
                 port=pb.portno):
        """Initialize a service.

        @param serviceName: The name of the PB service.  If you make
        more than one service, each must have a unique name.

        @param userpass: If a (username, password) tuple is supplied
        here, an account will be created for it.

        @param port: The number of the TCP port I should listen on.
        """
        components.Componentized.__init__(self)

        if userpass is not None:
            try:
                user, passwd = userpass
            except (ValueError, TypeError):
                raise TypeError("userpass must be a 2-tuple"
                                " (username, password)")
            self.userpass = userpass

        self.portal = portal.Portal(portal.IRealm(self))
        self.portal.registerChecker(checkers.AllowAnonymousAccess())
        # XXX: FIXME: Use a FilePasswordDB instead!
        # (Actually, to some degree, this is valid, since the current
        # PB login only lets you use IUsernamePassword, even though
        # I would like to accept IAnonymous most times.)
        fakeUsernamePassword = checkers.AllowAnonymousAccess()
        fakeUsernamePassword.credentialInterfaces = (
            checkers.credentials.IUsernamePassword,
            checkers.credentials.IUsernameHashedPassword)
        self.portal.registerChecker(fakeUsernamePassword)
        self.netservice = strports.service(str(port),
                                           pb.PBServerFactory(self.portal))
        self.netservice.setName(serviceName)
        self.netservice.setServiceParent(IServiceCollection(self))

    def __call__(self, root, files, message, user):
        for c in portal.IRealm(self).connectedClients():
            c.notify(root, files, message, user)

    def unload(self):
        # disownServiceParent results in the service being stopped,
        # so I don't need to IService(self).stopService()
        return defer.maybeDeferred(IService(self).disownServiceParent)

components.registerAdapter(_PBRealm, PBService, portal.IRealm)
components.registerAdapter(_PBService, PBService, IService)
components.registerAdapter(_PBService, PBService, IServiceCollection)

# Install in 'actions' namespace so it gets picked up by configatron.
actions.PBService = PBService


# Local Variables:
# elide-head-headers-to-hide:(("<license" . "</license>"))
# End:
