# -*- Python -*-
# $Id: pb.py,v 1.12 2004/02/19 10:19:18 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, August 2002
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""Base classes handy for use with PB clients.
"""

from twisted.internet import protocol, reactor
from twisted.python import log
from twisted.spread import pb

import random

from cvstoys import common

class PerspectiveInUse(Exception):
    """Raised when a client requests a perspective already connected to another.
    """
    pass

class Backoff:
    # Code duplication with protocol.ReconnectingClientFactory
    maxDelay = 3600
    initialDelay = 1.0

    # Note: These highly sensitive factors have been precisely measured by
    # the National Institute of Science and Technology.  Take extereme care
    # in altering them, or you may damage your Internet!
    factor = 2.7182818284590451 # (math.e)
    # Phi = 1.6180339887498948 # (Phi is acceptable for use as a
    # factor if e is too large for your application.)
    jitter = 0.11962656492 # molar Planck constant times c, Jule meter/mole

    delay = initialDelay
    count = 0
    maxRetries = None
    _callID = None
    perspective = None

    def next(self):
        self.count += 1
        self.delay = min(self.delay * self.factor, self.maxDelay)
        if self.jitter:
            self.delay = random.normalvariate(self.delay,
                                              self.delay * self.jitter)
        return self.delay

    def reset(self):
        self.delay = self.initialDelay
        self.count = 0

    def callLater(self, callable, *args, **kw):
        self.next()
        return reactor.callLater(self.delay, callable, *args, **kw)

    def __repr__(self):
        return "<%s at %x, count %d, value %d>" % (
            self.__class__, id(self), self.count, self.delay)

    def __str__(self):
        return "#%d:%ds" % (self.count, self.delay)


# Adopted from buildbot.pbutil.  Thank you Brian Warner.
class ReconnectingPBClientFactory(pb.PBClientFactory,
                                  protocol.ReconnectingClientFactory):
    """Reconnecting client factory for PB brokers.

    Like PBClientFactory, but if the connection fails or is lost, the factory
    will attempt to reconnect.

    Instead of using f.getRootObject (which gives a Deferred that can only
    be fired once), override the gotRootObject method.

    Instead of using the newcred f.login (which is also one-shot), call
    f.startLogin() with the credentials and client, and override the
    gotPerspective method.

    Instead of using the oldcred f.getPerspective (also one-shot), call
    f.startGettingPerspective() with the same arguments, and override
    gotPerspective.

    gotRootObject and gotPerspective will be called each time the object is
    received (once per successful connection attempt). You will probably want
    to use obj.notifyOnDisconnect to find out when the connection is lost.

    If an authorization error occurs, failedToGetPerspective() will be
    invoked.

    To use me, subclass, then hand an instance to a connector (like
    TCPClient).
    """

    def __init__(self):
        pb.PBClientFactory.__init__(self)
        self._doingLogin = False
        self._doingGetPerspective = False

    def clientConnectionFailed(self, connector, reason):
        pb.PBClientFactory.clientConnectionFailed(self, connector, reason)
        RCF = protocol.ReconnectingClientFactory
        RCF.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, reason):
        pb.PBClientFactory.clientConnectionLost(self, connector, reason,
                                             reconnecting=True)
        RCF = protocol.ReconnectingClientFactory
        RCF.clientConnectionLost(self, connector, reason)

    def clientConnectionMade(self, broker):
        self.resetDelay()
        pb.PBClientFactory.clientConnectionMade(self, broker)
        if self._doingLogin:
            self.doLogin(self._root)
        if self._doingGetPerspective:
            self.doGetPerspective(self._root)
        self.gotRootObject(self._root)

    def __getstate__(self):
        # this should get folded into ReconnectingClientFactory
        d = self.__dict__.copy()
        d['connector'] = None
        d['_callID'] = None
        return d

    # oldcred methods

    def getPerspective(self, *args):
        raise RuntimeError, "getPerspective is one-shot: use startGettingPerspective instead"

    def startGettingPerspective(self, username, password, serviceName,
                                perspectiveName=None, client=None):
        self._doingGetPerspective = True
        if perspectiveName == None:
            perspectiveName = username
        self._oldcredArgs = (username, password, serviceName,
                             perspectiveName, client)

    def doGetPerspective(self, root):
        # oldcred getPerspective()
        (username, password,
         serviceName, perspectiveName, client) = self._oldcredArgs
        d = self._cbAuthIdentity(root, username, password)
        d.addCallback(self._cbGetPerspective,
                      serviceName, perspectiveName, client)
        d.addCallbacks(self.gotPerspective, self.failedToGetPerspective)


    # newcred methods

    def login(self, *args):
        raise RuntimeError, "login is one-shot: use startLogin instead"

    def startLogin(self, credentials, client=None):
        self._credentials = credentials
        self._client = client
        self._doingLogin = True

    def doLogin(self, root):
        # newcred login()
        d = self._cbSendUsername(root, self._credentials.username,
                                 self._credentials.password, self._client)
        d.addCallbacks(self.gotPerspective, self.failedToGetPerspective)


    # methods to override

    def gotPerspective(self, perspective):
        """The remote avatar or perspective (obtained each time this factory
        connects) is now available."""
        pass

    def gotRootObject(self, root):
        """The remote root object (obtained each time this factory connects)
        is now available. This method will be called each time the connection
        is established and the object reference is retrieved."""
        pass

    def failedToGetPerspective(self, why):
        self.stopTrying() # logging in harder won't help
        log.err(why)

class ReconnectingPB(pb.Referenceable):
    """Reconnecting PB client.

    XXX: This duplicates a lot of code with protocol.ReconnectingClientFactory.

    @ivar perspective: Will be a
        L{RemoteReference<twisted.spread.pb.RemoteReference>} to my
        L{Perspective<twisted.spread.pb.Perspective>} when I am connected,
        and C{None} when I am not.
    @type perspective:
        L{RemoteReference<twisted.spread.pb.RemoteReference>} of
        L{Perspective<twisted.spread.pb.Perspective>}
    """
    retries = 0
    maxRetries = None
    _callID = None
    perspective = None
    _backoff = None

    def __init__(self, host, port, user, passwd, serviceName):
        self.host = host
        self.port = port
        self.user = user
        self.passwd = passwd
        self.serviceName = serviceName

    def retry(self):
        """Have this connector connect again, after a suitable delay.
        """
        if self._backoff is None:
            self._backoff = Backoff()

        self.retries += 1
        if self.maxRetries is not None and (self.retries > self.maxRetries):
            log.msg("Abandoning %s after %d connection retries." %
                    (self, self.retries))
            return

        self._callID = self._backoff.callLater(self.connect)
        log.msg("%s will retry in %d seconds" % (self, self._backoff.delay,))

    def stopTrying(self):
        if self._callID:
            self._callID.cancel()

    def connect(self):
        pb.connect(self.host, self.port, self.user, self.passwd,
                   self.serviceName,
                   client=self).addCallbacks(self.connected, self.failed)

    def connected(self, perspective):
        log.msg("%s successfully connected." % (self,))
        self._callID = None
        if self._backoff is not None:
            self._backoff.reset()
        self.perspective = perspective
        perspective.notifyOnDisconnect(self.disconnected)

    def disconnected(self, unused_ref):
        log.msg( "%s disconnected." % (self,))
        self.perspective = None
        self.retry()

    def failed(self, reason):
        log.msg( "%s is sad because %s" % (self, reason.value))
        self.retry()

    def __getstate__(self):
        state = self.__dict__.copy()
        for v in 'perspective', '_callID', '_backoff':
            try:
                del state[v]
            except KeyError:
                pass
        return state



# Local Variables:
# elide-head-headers-to-hide:(("<license" . "</license>"))
# End:
