#$Id: words.py,v 1.25 2004/03/05 06:30:27 acapnotic Exp $

"""Deliver notifications through a chat server, such as Words or IRC.

TOC (i.e. AIM) support is theoretically possible as well (it's supported
by the underlying twisted.im package), but has never been tested.

Stability: Released versions of this code are functional.  However,
the internal structure is highly unstable, following L{twisted.im}
changes and my own evolving ideas about how to best organize it.  If
you're patching or extending this code, please stay in close contact
with the maintainer.
"""

__version__ = '$Revision: 1.25 $'[11:-2]

from __future__ import nested_scopes

# TODO:
#
#   Deal with getting Kicked.  Somehow.
#

from twisted.python.compat import *

from twisted.application import internet
from twisted.application.service import MultiService
from twisted.internet import reactor
from twisted.internet import app, defer
from twisted.persisted import styles
from twisted.python import components, log, reflect, usage
from twisted.spread import pb
from twisted.im import baseaccount, basechat, basesupport
from twisted.im import interfaces as imInterfaces
from twisted.im.locals import OfflineError
from os import path

# XXX: Having this hardcoded isn't very cool.  We should automatically adapt
# to any implementation of twisted.im.
from twisted.im import ircsupport, pbsupport, tocsupport

from cvstoys import common, config, interfaces
from cvstoys.actions import actions

# siblings
import pb as pbcvs

MAXFILES = 7

class ICrier(components.Interface):
    def notify(self, root, files, message, user):
        pass

    def goodbye(self, message):
        pass

    def notifieeConnected(self, notifiee):
        pass

    def notifieeDisconnected(self, notifiee, reason):
        pass

    def notifieeConnecting(self, notifiee):
        pass


class Notifiee(pb.Referenceable):
    __implements__ = (interfaces.INotifieeClient,)

    crier = None
    # FIXME: What if there's more than one avatar?
    avatar = None
    def remote_notify(self, root, files, message, user):
        self.crier.notify(root, files, message, user)

    def remote_goodbye(self, message):
        log.msg("Disconnecting at server's request: %s" % (message,))
        self.crier.goodbye(message)
        self.avatar.broker.transport.loseConnection()

    def __getstate__(self, state=None):
        if state is None:
            state = self.__dict__.copy()
        try:
            del state['avatar']
        except KeyError:
            pass
        return state


class NotifieeFactory(pbcvs.ReconnectingPBClientFactory):
    def gotPerspective(self, p):
        self.notifiee.crier.notifieeConnected(self)
        self.notifiee.avatar = p
        return pbcvs.ReconnectingPBClientFactory.gotPerspective(self, p)

    def clientConnectionLost(self, connector, reason):
        self.notifiee.crier.notifieeDisconnected(self, reason)
        return pbcvs.ReconnectingPBClientFactory.clientConnectionLost(
            self, connector, reason)

    def startLogin(self, credentials, client):
        self.notifiee = client
        if self.notifiee.crier:
            self.notifiee.crier.notifieeConnecting(self)
        return pbcvs.ReconnectingPBClientFactory.startLogin(
            self, credentials, client)


nop = lambda *x, **y: None

class Conversation(basechat.Conversation):
    __implements__ = (imInterfaces.IConversation,)

    def __init__(self, person):
        self.person = person

    def showMessage(self, text, metadata):
        if metadata:
            m = '<%s>' % (metadata,)
        else:
            m = ''
        log.msg("%s%s: %r" % (self.person, m, text))

    show = hide = nop


class GroupConversation(basechat.GroupConversation):
    __implements__ = (imInterfaces.IGroupConversation,)
    def __init__(self, group):
        self.group = group

    show = hide = showGroupMessage = setTopic = nop
    memberJoined = memberLeft = memberChangedNick = nop

class GroupMessage(components.Adapter, actions.CommitAction, reflect.Accessor):
    __implements__ = (interfaces.ICommitAction,)

    def __call__(self, root, files, message, user):
        try:
            self.original.sendGroupMessage(makeMessage(root, files,
                                                       message, user))
        except OfflineError, e:
            return defer.fail(e)
        # Technically speaking, sendGroupMessage oughta return Deferred and
        # callback when the message has been confirmed of delivery without
        # error, but that's not in the twisted.im sendMessage interface.
        return defer.succeed(None)

    def sendText(self, text):
        try:
            self.original.sendGroupMessage(text)
        except OfflineError:
            pass

    def emote(self, text):
        try:
            self.original.sendGroupMessage(text, {'style': 'emote'})
        except OfflineError:
            pass

    def join(self):
        self.original.join()

    def get_account(self):
        return self.original.account

# Can't actually adapt one interface to another right now.
#
# components.registerAdapter(GroupMessage,
#                           imInterfaces.IGroup, interfaces.ICommitAction)

for groupimpl in (ircsupport.IRCGroup,
                  pbsupport.TwistedWordsGroup,
                  tocsupport.TOCGroup):
    components.registerAdapter(GroupMessage, groupimpl,
                               interfaces.ICommitAction)

class TimCrier(styles.Versioned):
    __implements__ = (ICrier,)
    VERBOSE_RECONNECTS = False
    persistenceVersion = 1

    config = None

    def __init__(self):
        self.groups = []

    def notify(self, root, files, message, user):
        for action in self.config.mapFilesToActions(root, files):
            action['action'](root, action['files'], message, user)

    def goodbye(self, message):
        for group in self.groups:
            group.sendText("Server said goodbye: %s" % (message,))

    def notifieeConnected(self, unused_notifiee):
        for group in self.groups:
            group.emote("connected to CVS notification service.")

    def notifieeDisconnected(self, unused_notifiee, reason):
        for group in self.groups:
            group.emote("disconnected [%s]" % (reason,))

    def notifieeConnecting(self, unused_notifiee):
        if self.VERBOSE_RECONNECTS:
            for group in self.groups:
                group.emote("is connecting...")

    def connect(self, chatui):
        accounts = []
        for group in self.groups:
            if group.account not in accounts:
                accounts.append(group.account)
        log.msg("%s connecting %d clients for %d groups." %
                (self, len(accounts), len(self.groups)))
        for account in accounts:
            d = account.logOn(chatui)

            # XXX: d.addCallback(self.loggedOn) happens in
            # chatui.registerAccountClient instead of here.  Why?
            # Because if we add a callback here, it'll only get fired
            # if logOn works the first time, not if it succeeds only
            # on a retry.  More reasons for cleaning up t.im
            # reconnecting.

            d.addErrback(chatui._ebReconnect, account)

    def loggedOn(self, client):
        for g in self.groups:
            if g.account is client.account:
                g.join()

    def setGroups(self, groups):
        # log.msg("%s setGroups: %s" % (self, groups))
        self.groups[:] = groups

class FactoryCache:
    def __init__(self, factory):
        self.factory = factory
        self._cache = {}

    def __call__(self, *a, **kw):
        key = a
        if not self._cache.has_key(key):
            self._cache[key] = self.factory(*a, **kw)
        return self._cache[key]

    def clear(self):
        self._cache.clear()

class Tim(basechat.ChatUI):
    __implements__ = (imInterfaces.IChatUI,)

    _conversationFactory = FactoryCache(Conversation)
    _groupConversationFactory = FactoryCache(GroupConversation)

    # This attribute, like most that look they're in the wrong place, is kept
    # around for calling crier.loggedOn in the client-reconnecter's callback.
    crier = None

    def __init__(self):
        basechat.ChatUI.__init__(self)
        self._reconnectors = {}

    def registerAccountClient(self, client):
        r = basechat.ChatUI.registerAccountClient(self, client)
        if self._reconnectors.has_key(client.account):
            self._reconnectors[client.account].reset()
        self.crier.loggedOn(client)
        return r

    def unregisterAccountClient(self, client):
        retval = basechat.ChatUI.unregisterAccountClient(self, client)
        self._reconnect(client.account)
        log.msg("%s logged off" % (client,))
        return retval

    # XXX: I don't like having this reconnect stuff in ChatUI here; it
    # seems like an AccountManager job to me.

    def _reconnect(self, account):
        if not self._reconnectors.has_key(account):
            self._reconnectors[account] = pbcvs.Backoff()
        backoff = self._reconnectors[account]
        def connect():
            account.logOn(self).addErrback(self._ebReconnect, account)
        backoff.callLater(connect)
        log.msg("%s reconnecting in %d" % (account, backoff.delay))

    def _ebReconnect(self, reason, account):
        log.msg("%s reconnect failed: %s" % (account, reason.value))
        self._reconnect(account)

    def getConversation(self, person, unusedClass=None, stayHidden=0):
        return self._conversationFactory(person)

    def getGroupConversation(self, group, unusedClass=None, stayHidden=0):
        return self._groupConversationFactory(group)

    def __getstate__(self):
        state = self.__dict__.copy()
        if state.has_key("_reconnectors"):
            state['_reconnectors'] = {}
        return state


def makeMessage(root, files, message, user):
    new = 0
    removed = 0
    changed = 0
    directories = {}
    if (len(files) == 1) and isinstance(files[0], common.Directory):
        m = "Commit from %s: New directory %s/%s" % (user,
                                                     files[0].pathname,
                                                     files[0].filename)
        return m

    for f in files:
        if f.r1 == (None,):
            new += 1
        elif f.r2 == (None,):
            removed += 1
        else:
            changed += 1
        directories[f.pathname] = 'Set'

    l = []
    if removed:
        l.append("removed %d" % (removed,))
    if new:
        l.append("added %d" % (new,))
    if changed:
        l.append("changed %d" % (changed,))

    l = ", ".join(l)

    fs = [f.filename for f in files]
    if len(fs) > (MAXFILES + 1):
        fs[MAXFILES:] = ["%d more..." % (len(fs) - MAXFILES,)]
    fs = ", ".join(fs)

    # TODO: if the only file changed in a particular directory is ChangeLog,
    # then drop that directory from the list.
    if len(directories) > 1:
        commonpath = actions.commonprefix(directories.keys())
        ds = "%d subdirs of %s" % (len(directories), commonpath)
    elif directories:
        ds = directories.keys()[0]

    m = ("Commit from %(who)s (%(numbers)s) in %(dirs)s: "
         "\"%(msg)s\" %(files)s" % {'who': user,
                                    'numbers': l,
                                    'dirs': ds,
                                    'msg': message.split('\n', 1)[0],
                                    'files': fs})
    return m


class Service(MultiService):
    _config = None
    configFilePath = None

    def __init__(self, notifiee):
        MultiService.__init__(self)

        self.chatUI = Tim()
        self.crier = TimCrier()

        # FIXME: Componentize!
        notifiee.crier = self.chatUI.crier = self.crier

    def startService(self):
        if self.configFilePath:
            self._config = Configuration()
            self._config.load(self.configFilePath)

            self.crier.config = self._config
            self.crier.setGroups(self._config.getGroups())

        MultiService.startService(self)

        self.crier.connect(self.chatUI)


class Options(usage.Options):
    optParameters = [
        ('config', 'f', None, "Configuration file."),
        ('cvs-host', 'H', 'localhost', "FreshCVS host."),
        ('cvs-port', 'P', 7878, "FreshCVS PB port."),
        ('cvs-user', 'U', 'words', "FreshCVS username."),
        ('cvs-pass', 'W', 'notifiee', "FreshCVS password."),
        # vestigial
        ('cvs-service', 'S', 'cvstoys.notify', "FreshCVS service."),
        ]

class Configuration(config.Configuration):
    """Chat account configuration.

    Defines what chat accounts to connect to, and what messages to deliver
    to which groups.

    Tips: The I{autoLogin} hint in the account constructor is not used.
    """

    _builtinDict = {
        'IRCAccount': ircsupport.IRCAccount,
        'PBAccount': pbsupport.PBAccount,
        'TOCAccount': tocsupport.TOCAccount,
        }

    def getGroups(self):
        actions = self.getActions()
        groups = [g for g in actions if isinstance(g, GroupMessage)]
        log.msg("%s.getGroups: of %d actions, %d are groups." % (self,
                                                                 len(actions),
                                                                 len(groups)))
        return groups


def makeService(options):
    from twisted.cred import credentials

    # for each notifiee:
    notifiee = Notifiee()
    factory = NotifieeFactory()
    mycreds = []
    #if options['cvs-user']:
    mycreds.append(credentials.UsernamePassword(options['cvs-user'],
                                                options['cvs-pass']))
    #else:
    # mycreds.append(credentials.Anonymous())
    factory.startLogin(mycreds[0], notifiee)
    # FIXME: no strports.client (Twisted issue 511)
    c = internet.TCPClient(options['cvs-host'], int(options['cvs-port']),
                           factory)
    # next notifiee

    thisServ = Service(notifiee)
    thisServ.configFilePath = options['config']
    c.setServiceParent(thisServ)

    return thisServ
