# -*- Python -*-
# $Id: config.py,v 1.25 2004/03/05 08:20:24 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, October 2001
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""I handle the ins and outs of configuration for the Fresh CVS daemon.

About the configuration file:

By default I look for it in $CVSROOT/CVSROOT/freshCfg and expect to be
able to run it as Python.  The class ConfigurationSet and any action
classes from actions.py will be in its global namespace.

Example::

    | set1 = ConfigurationSet([
    |     (None, '^Twisted', '\.py$', MailNotification(['commits@pythonic.org',
    |                                                   'grease@pokey.fi']))
    |     (None, '^Teud', None, MailNotification(['teud-dev@pythonic.org']))
    |     ])
    |
    | set2 = ConfigurationSet([
    |     (None, '.*twisted.*', None, None),
    |     ('/usr/local/share/reposA', '^murdock', None,
    |      CheckOut('/home/A-team/public_html/')),
    |     ('/usr/local/share/reposA', None, None,
    |      CheckOut('/usr/local/src/A/')),
    |     ('/usr/local/share/reposB', None, None,
    |      CheckOut('/usr/local/src/B/')),
    |     (None, None, None, DebugAction(\"Funny repository!\"))
    |     ])
"""

__version__ = '$Revision: 1.25 $'[11:-2]

import re, sys
from os import path
from types import ClassType

from twisted.application.service import IService
from twisted.internet import defer
from twisted.persisted import styles
from twisted.python import components, log, reflect

from cvstoys.actions import actions
from cvstoys.actions import pb as pbAction
import interfaces

class Configuration:
    _builtinDict = {} # populated later
    def __init__(self, *args, **kw):
        self.configSets = []
        self._builtinDict = {}
        self._localDict = {}

        self.adapterRegistry = {}
        self.createAdapters(*args, **kw)

    def createAdapters(self, *args, **kw):
        pass

    def registerAdapter(self, adapterFactory, origClass, iface):
        self.adapterRegistry[(origClass, iface)] = adapterFactory

    def getAdapter(self, obj, iface):
        klass = obj.__class__
        adapterFactory = self.adapterRegistry.get((klass, iface))
        if adapterFactory is not None:
            return adapterFactory(obj)
        else:
            return components.getAdapter(obj, iface)

    def load(self, filepath):
        """What do you *think* configuration.load does?
        """

        file_ = open(filepath, 'r')

        config_local_dict = self._localDict.copy()

        # exec adds things to the globals dict that aren't pickleable, so
        # make sure we pass a local copy to exec, not anything that might
        # get pickled.
        builtinDict = self.__dict__.get('_builtinDict', {}).copy()
        reflect.accumulateClassDict(self.__class__,
                                    "_builtinDict", builtinDict)

        try:
            exec file_ in builtinDict, config_local_dict
        finally:
            file_.close()

        configSets = []

        if config_local_dict.has_key("allSets"):
            configSets.extend(config_local_dict["allSets"])
        else:
            for (name, obj) in config_local_dict.items():
                if isinstance(obj, ConfigurationSet):
                    obj.name = name
                    configSets.append(obj)

        log.msg("%s config sets found" % (len(configSets),))

        for cs in configSets:
            cs.process(self)

        # XXX: This re-sets the set of configs instead of adding to it.  That
        # is bad practice if we have more than one configuration file.
        self.setConfigs(configSets)

        return config_local_dict

    def unload(self):
        # Retire old configuration.
        dl = []
        for cs in self.configSets:
            dl.append(cs.unload())
        d = defer.DeferredList(dl)
        return d

    def setConfigs(self, configSets):
        self.configSets[:] = configSets

    def getActions(self):
        # I wish I were an iterator.
        l = []
        for cfgset in self.configSets:
            l.extend(cfgset.getActions())
        return l

    def mapFilesToActions(self, root, files):
        l = []
        for cfgset in self.configSets:
            l.extend(cfgset.mapFilesToActions(root, files))
        return l


class FresherConfiguration(Configuration):
    _builtinDict = {} # populated later
    def __init__(self, root, serviceParent):
        # Urgh.  The whole idea behind breaking out createAdapters into another
        # method is that you wouldn't need to override __init__ anymore, right?
        # But apparently we do other things too:
        #  - Store instance attributes for later use in load().
        #  - Play with namespace dicts.
        #
        # Really, it's the variable signature that's bugging me.
        # Configuration.__init__ takes *a, **kw and passes them
        # on to createAdapters, so you can write a createAdapters
        # with any signature you like.  And that makes sense if you
        # just kind of pretend Configuration(foo, bar) just calls
        # myConfiguration.createAdapters(foo, bar).  But in this case, we've
        # got FresherConfiguration.__init__ taking one thing and then passing
        # something else on to Configuration.__init__, not because that's the
        # signature that Configuration.__init__ wants, but because it's the
        # signature FresherConfiguration.createAdapters wants.  Smelly!

        Configuration.__init__(self, serviceParent)
        self.serviceParent = serviceParent
        builtinSet = ConfigurationSet([
            (root, '^CVSROOT$', 'freshCfg', ReloadAction(self)),
            ])
        self._localDict['_fresherBuiltinSet'] = builtinSet

    def createAdapters(self, serviceParent):
        # Woo abuse of the adapter registry!
        # and _setServiceParent was a locally defined function, but pickle
        # couldn't deal with that.
        self.registerAdapter(_setServiceParent(serviceParent),
                             pbAction.PBService,
                             interfaces.ICommitAction)

    # FIXME: pbPortNumber backwards compatibility

class _setServiceParent:
    def __init__(self, serviceParent):
        self.serviceParent = serviceParent

    def __call__(self, orphanService):
        IService(orphanService).setServiceParent(self.serviceParent)
        return orphanService

class ConfigurationSet(styles.Versioned):
    """One collection of patterns and their resulting actions.

    A configuration set consists of a list of
    (root, path_regex, file_regex, action) tuples.

    This is an ordered list, and the *first* matching line wins.
    If you wish multiple actions to occour for a single event,
    use multiple configuration sets.

    'root' is not a regex, but will match any if it is None.
    An action of None will be a no-op.
    """

    persistenceVersion = 1

    # XXX: we should probably provide an option of this without 'root'
    # arg and use the root of whatever CVSROOT we're pulling this
    # configuration from.  The current interface could still be used for
    # a global configuration file.

    # XXX: If we wanted to be really badass, we'd implement the
    # ConfigurationSet *as* an action, and thus have them nestable.
    # (aiee!)

    _matchAll = re.compile('')
    name = None
    unprocessed = None
    configuration = None

    def __init__(self, configlist=None):
        self.configs = []

        if configlist:
            self.unprocessed = configlist

    def process(self, configuration):
        self.configuration = configuration
        if self.unprocessed:
            self.addConfigs(self.unprocessed)
            self.unprocessed[:] = []

    def addConfigs(self, configlist):
        for i in range(len(configlist)):
            try:
                cfg = self._importLine(configlist[i])
            except (TypeError, ValueError), e:
                # An attempt to know what configuration set we are, but
                # probably futile since we won't know our name until
                # *after* we've been initialized.  Another reason why
                # I should be a metaclass.

                s = "Error processing configuration line #%d of %s: %s" \
                    % (i + 1, self.name or '(unnamed)', e.args[0])
                e.args = list(e.args)
                e.args[0] = s
                # XXX: This doesn't get all the traceback you need.
                raise e
            else:
                self.configs.append(cfg)

    def _importLine(self, line):
        (root, path_reString, file_reString, action) = line

        if action:
            action = self.configuration.getAdapter(action,
                                                   interfaces.ICommitAction)

        reStrings = {'path': path_reString,
                     'file': file_reString}

        res = {}
        for (name, rString) in reStrings.items():
            if rString is None:
                res[name] = self._matchAll
            else:
                try:
                    res[name] = re.compile(rString)
                except (TypeError, ValueError), e:
                    s = "Error compiling %s regex '%s': %s" \
                        % (name, rString, e.args[0])
                    e.args = list(e.args)
                    e.args[0] = s
                    raise e

        cfg = {'root': root,
               'path_re': res['path'],
               'file_re': res['file'],
               'action': action
               }

        return cfg

    def getActions(self):
        """List all non-None actions I have."""
        actions = [cfg['action'] for cfg in self.configs
                   if cfg['action'] is not None]
        return actions

    def unload(self):
        """Tell all my actions to unload themselves."""
        dl = []
        for cfg in self.configs:
            action = cfg['action']
            if action is None:
                continue
            if not components.implements(action, interfaces.ICommitAction):
                action = action.getComponent(interfaces.ICommitAction)
            dl.append(action.unload())
            if not isinstance(dl[-1], defer.Deferred):
                raise TypeError, "action %s returned %r for unload" % (
                    action, dl[-1])
        return defer.DeferredList(dl)

    def mapFilesToActions(self, root, files):
        """Given files, I will respond with which actions are configured to take them.

        Returns a list of dictionaries:
        [{'action': action1,
          'files': [fileA, fileB, ...]},
         {'action': action2,
          'files': [fileC, fileD, ...]}
          ]
        """

        actions = {}

        for file_ in files:
            pathname = file_.pathname
            filename = file_.filename

            for cfg in self.configs:
                if cfg['root'] and (root != cfg['root']):
                    continue
                if cfg['path_re'].search(pathname) is None:
                    continue
                if cfg['file_re'].search(filename) is None:
                    continue

                action = cfg['action']
                if ((action is not None)
                    and not components.implements(action,
                                                  interfaces.ICommitAction)):
                    action = action.getComponent(interfaces.ICommitAction)

                if not actions.has_key(action):
                    actions[action] = []

                actions[action].append(file_)
                # configuration matched, check no more for this file.
                break

        actionList = []
        for (action, files) in actions.items():
            actionList.append({'action': action,
                               'files': files})

        return actionList


class ReloadAction(actions.CommitAction):
    """Re-load a Configuration file.

    @cvar delaySecs: Wait this many seconds to give cvs a chance to
    \"[rebuild its] administrative file database.\"
    @type delaySecs: non-negative number
    """

    delaySecs = 5
    def __init__(self, config):
        """Initialize with a Configuration to take action on.

        @param config: The object I will call load() on.
        @type config: L{Configuration}
        """
        self.config = config

    def __call__(self, root, files, message, user):
        from twisted.internet import reactor
        assert files
        if len(files) > 1:
            log.msg("Warning: Executing a ReloadAction with more than "
                    "one file. (Files: %s)" % (files,))
        _deferredResult = defer.Deferred()
        reactor.callLater(self.delaySecs, self._doLoad,
                          root, files, message, user, _deferredResult)
        return _deferredResult

    def _doLoad(self, root, files, message, user, _deferredResult):
        cfgfile = path.join(root, files[0].pathname, files[0].filename)
        log.msg("Loading %s version %s into %r." % (cfgfile, files[0].r2,
                                                    self.config))
        log.msg("%s commit message for %s:" % (user, cfgfile))
        log.msg(message)

        # When reloading, unloading must happen before loading.
        _deferredResult.addCallback(lambda unused: self.config.unload())
        _deferredResult.addCallback(lambda unused: self.config.load(cfgfile))
        _deferredResult.callback(None)


# Silly documentation stuff.  TODO: Replace me with epydocness.

import inspect
def make_docs(f=None):
    if f is None:
        f = sys.stdout
    f.write(inspect.getdoc(sys.modules[__name__]))

    f.write("\n\nConfigurationSet\n")
    f.write(inspect.getdoc(ConfigurationSet))
    docinit(ConfigurationSet, f)

    f.write('\n----------\n')

    f.write("\nThe pre-defined actions you may use are:\n\n")

    for k, v in FresherConfiguration._builtinDict.items():
        if k in ('CommitAction', 'ConfigurationSet'):
            continue

        f.write("%s\n" % (k,))
        if v.__doc__:
            f.write(inspect.getdoc(v).rstrip() + '\n')
        else:
            f.write("No doc.\n")

        docinit(v, f)

        f.write("\n----\n\n")

def docinit(klass, f):
    if hasattr(klass, '__init__'):
        init = klass.__init__
        argspec = inspect.getargspec(init.im_func)
        # eat 'self'
        argspec[0].pop(0)
        if argspec[0]:
            callspec = inspect.formatargspec(*argspec)
            f.write("\nUsage: %s%s\n" % (klass.__name__, callspec))
            if init.__doc__:
                f.write(inspect.getdoc(init).rstrip() + '\n')


# Build the global namespace the configuration files will be loaded in.
# (we do this at the end so any classes we might have need of above
#  are defined.)
import cvstoys.actions.mail
for (_name, _object) in actions.__dict__.items():
    if isinstance(_object, ClassType) \
       and issubclass(_object, actions.CommitAction):
        FresherConfiguration._builtinDict[_name] = _object

Configuration._builtinDict.update({
    "ConfigurationSet": ConfigurationSet,
    })
