# -*- Python -*-
# $Id: fresher.py,v 1.39 2004/03/04 07:37:49 acapnotic Exp $
#
# By Kevin Turner <acapnotic@twistedmatrix.com>, October 2001
# <license common_name="BSD license">
# Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the
# following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the
#    above copyright notice, this list of conditions and
#    the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#
#    3. The name of the author may not be used to endorse
#    or promote products derived from this software without
#    specific prior written permission.
#
# <!-- I am not happy with the person who makes it necessary
# to put something like the paragraph which follows on every
# header file and coffee cup. -->
#
# This software is provided by the author "as is" and any express
# or implied warranties, including, but not limited to, the implied
# warranties of merchantability and fitness for a particular purpose are
# disclaimed. In no event shall the author be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including, but not limited to, procurement of substitute goods or
# services; loss of use, data, or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort (including negligence or otherwise) arising
# in any way out of the use of this software, even if advised of the
# possibility of such damage.
# </license>

"""The Fresh CVS daemon.

I am the daemon which hangs around waiting for news from loginfo, and
then I dispatch actions as I am configured to do.
"""

__version__ = '$Revision: 1.39 $'[11:-2]

from twisted.python.compat import *

from twisted.spread import pb
from twisted.application import internet, strports
from twisted.application.service import MultiService, IServiceCollection
from twisted.internet import reactor
from twisted.persisted import styles
from twisted.python import components, log, usage, util

import errno, os, string, sys
from os import path

import pidutils
import common, config


class Options(usage.Options):
    optParameters = [('manhole-port', None,
                      # default to portno +2 because pbservice will default
                      # to portno.
                      pb.portno + 2, "manhole port.")]

    def postOptions(self):
        # Set default root path.
        # TODO: give usage.Options a better way to specify defaults
        # for options with opt_() methods.
        if not self.opts.get('root', None):
            if (os.environ.has_key('CVSROOT')
                and os.environ['CVSROOT'][0] != ':'):

                self.opts['root'] = os.environ['CVSROOT']
            else:
                self.opts['root'] = os.getcwd()

        if not self.opts.get('pipe', None):
            self.opts['pipe'] = path.join(self.opts['root'], 'CVSROOT',
                                          common.SOCKETNAME)

    def opt_pipe(self, value):
        """The unix socket to listen on."""
        directory = path.dirname(value)
        if not path.exists(directory):
            raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), directory)
        if not path.isdir(directory):
            raise IOError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), directory)
        if not os.access(directory, os.R_OK | os.W_OK | os.X_OK):
            raise IOError(errno.EACCES, os.strerror(errno.EACCES), directory)
        self.opts['pipe'] = value

    def opt_root(self, value):
        """The directory where the configuration file lives.
        """
        if not path.exists(value):
            raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), value)
        if not path.isdir(value):
            raise IOError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), value)
        # Actually, if the pipe goes somewhere else, I don't need write
        # permissions to this directory.
        # if not os.access(value, os.R_OK | os.W_OK | os.X_OK):
        if not os.access(value, os.X_OK):
            raise IOError(errno.EACCES, os.strerror(errno.EACCES), value)
        self.opts['root'] = path.normpath(path.abspath(value))

    def opt_manhole(self, password):
        """Password to the manhole. - will read from stdin or prompt.
        """
        # If standard input is a terminal, I prompt for a password and
        # confirm it.  Otherwise, I use the first line from standard
        # input, stripping off a trailing newline if there is one.
        if password in ('', '-'):
            self.opts['manhole'] = util.getPassword(confirm=True)
        else:
            self.opts['manhole'] = password

    opt_r = opt_root


class IRepositoryTable(components.Interface):
    # TODO: Fill this in.
    pass

class RepositoryTable(components.Adapter):
    """A table of all repositories.

    I am a table of all repositories we know about, keyed by their
    directory name (that which you would find in $CVSROOT).
    """

    __implements__ = IRepositoryTable,

    def __init__(self, service):
        components.Adapter.__init__(self, service)
        self.repositories = {}

    def addRepository(self, r):
        self.repositories[r.root] = r

    def listRepositories(self):
        return self.repositories.keys()

    def reloadConfig(self, root=None):
        if root is None:
            for r in self.repositories.values():
                r.reloadConfig()
        else:
            if not self.repositories.has_key(root):
                self.repositories[root] = Repository(root,
                                                     IServiceCollection(self))
            self.repositories[root].reloadConfig()

    # Utility method

    def addProcessEntry(self, root, pathname, files, pid, user, message):
        """Add a file to the process in the module in the repository in the repository table that Jack built.
        """
        if not self.repositories.has_key(root):
            self.repositories[root] = Repository(root,
                                                 IServiceCollection(self))

        return self.repositories[root].addProcessEntry(pathname, files,
                                                       pid, user, message)

    def addDirectoryForProcess(self, root, pathname, pid):
        if not self.repositories.has_key(root):
            self.repositories[root] = Repository(root,
                                                 IServiceCollection(self))

        return self.repositories[root].addDirectoryForProcess(pathname, pid)


class Repository:
    """A table of the modules in a repository.

    I am a table of all the modules we know about in a given repository.

    The string designating my repository's cvsroot is held in my
    member \"root\".
    """

    def __init__(self, root, service):
        self.root = root

        self.modules = {}
        self.config = config.FresherConfiguration(root, service)
        self.loadConfig()

    def listModules(self):
        return self.modules.keys()

    def reloadConfig(self):
        """Reload my configuration.

        Returns deferred.
        """
        d = self.config.unload()
        d.addCallback(lambda unused_result: self.loadConfig())
        return d

    def loadConfig(self):
        filepath = path.join(self.root, "CVSROOT", "freshCfg")
        log.msg("%s loading configuration from %s" % (self, filepath))
        self.config.load(filepath)

    def sweep(self):
        """Clean up unused entries.

        Iterates through all the module entries in my table, and
        deletes any which do not currently have processes.
        """
        for (m, pt) in self.modules.items()[:]:
            pt.sweep()
            if pt.isEmpty():
                del self.modules[m]

    def isEmpty(self):
        if len(self.modules) == 0:
            return True
        else:
            return False

    def getActionsFor(self, files):
        return self.config.mapFilesToActions(self.root, files)


    # Utility method

    def addProcessEntry(self, pathname, files, pid, user, message):
        """Add a file to the process in the module in the repository that Jack built.
        """
        moduleName = string.split(pathname, '/', 1)[0]

        if not self.modules.has_key(moduleName):
            self.modules[moduleName] = ProcessTable(moduleName, self)

        return self.modules[moduleName].addProcessEntry(pid, pathname,
                                                        files, user, message)

    def addDirectoryForProcess(self, pathname, pid):
        moduleName = string.split(pathname, '/', 1)[0]

        if not self.modules.has_key(moduleName):
            self.modules[moduleName] = ProcessTable(moduleName, self)

        return self.modules[moduleName].addDirectoryForProcess(pid, pathname)

    def __repr__(self):
        return "<%s %r>" % (self.__class__, self.root)

class ProcessTable:
    """A table of active PIDs and filenames active for a module."""

    def __init__(self, moduleName, repository):
        self.moduleName = moduleName
        self.repository = repository
        self.processes = {}

    def getRepository(self):
        return self.repository

    def listProcesses(self):
        return self.processes.keys()

    def sweep(self):
        for (pid, process) in self.processes.items()[:]:
            # XXX: Provide a non-pidutils alternative.
            if not pidutils.isDead(pid):
                continue

            # XXX: do stuff to files here

            del self.processes[pid]

    def isEmpty(self):
        if len(self.processes) == 0:
            return True
        else:
            return False

    def addProcessEntry(self, pid, pathname, files, user, message):
        """Add a file to the process in the module that Jack built.
        """
        if not self.processes.has_key(pid):
            self.processes[pid] = Process(pid, self)

        p = self.processes[pid]
        p.addFiles(pathname, files, user, message)

    def addDirectoryForProcess(self, pid, pathname):
        """Add a file to the process in the module that Jack built.
        """
        if not self.processes.has_key(pid):
            self.processes[pid] = Process(pid, self)

        p = self.processes[pid]
        p.addDirectory(pathname)

    def _removeProcess(self, pid):
        del self.processes[pid]

    def __getstate__(self):
        state = self.__dict__.copy()
        # These are Ephemeral
        state['processes'] = {}
        return state


class _Process(styles.Ephemeral):
    """I'm the CVS process you've been hearing about.

    Here is a list of all the files I've touched and my log message.
    I sit around here, checking my pulse every now and then, waiting
    for myself do die.  When I do, I pass along my belongings to the
    actions, and let them know they can start the party.

    @ivar files: A list of files updated by this process.
    @type files: A list of dicts with
    """

    message = None
    finished = False
    user = None

    EPSILON = 1.0

    def __init__(self, pid, module):
        # self.user = pidutils.owner(pid)
        self.pid = pid
        self.module = module

        self.files = []

    def getRepository(self):
        return self.module.getRepository()

    def addFiles(self, pathname, files, user, message):
        """Add a file to the process

        ...in...  phew! we're done.

        Appends files to the list of those touched by this process,
        and sets this process's log message.  It is assumed that a CVS
        process only gives us one unique log message, so it is only
        set once.
        """
        if self.finished:
            raise Exception("Process %s is already finished! "
                            "Can't add more files now." % (self.pid,))

        if user:
            if not self.user:
                self.user = user
            else:
                if user != self.user:
                    log.msg("Bug: How does this message come from user %s"
                            " when an earlier message to the same process"
                            " (%s) came from user %s?" % (repr(user),
                                                          self.pid,
                                                          repr(self.user),))

        if message and not self.message:
            # ASSUMPTION: only one message that matters per process
            # It is technically possible for the CVS client to send more than
            # one message per process, i.e. a different one for each directory,
            # but I haven't yet used a tool which actually does this.
            self.message = message

        for f in files:
            f.pathname = pathname
            self.files.append(f)

    def processComplete(self):
        actions = self.getRepository().getActionsFor(self.files)
        for action in actions:
            if action['action'] is None:
                continue
            reactor.callLater(self.EPSILON, action['action'],
                              self.getRepository().root,
                              action['files'],
                              self.message,
                              self.user)
        self.__removeSelf()

    def __removeSelf(self):
        """(internal) When my job is done...

        Remove the reference to me from my module's table.
        I'm finished and expect to be garbage collected.
        """

        self.module._removeProcess(self.pid)
        self.finished = True
        # and just for good measure,
        del self.module

    def addDirectory(self, unused):
        pass

class PollProcProcess(_Process):
    """Process variant that looks in /proc to see when it's done.

    Linux-specific and requires that I be hosted on the same machine
    as the CVS server.
    """

    def __init__(self, *a, **kw):
        _Process.__init__(self, *a, **kw)
        reactor.callLater(1, self.pulse)

    def pulse(self):
        if not pidutils.isDead(self.pid):
            reactor.callLater(1, self.pulse)
            return
        self.processComplete()


class DirectoryListProcess(_Process):
    """Process variant that depends on a list of directories visited.

    Requires a complete list of directories the process will visit,
    typically supplied by a commitinfo.  Declares itself done as soon
    as each directory has been visited.  If commitinfo lies or the
    process dies partway through, we're in trouble.
    """
    _loginfoBegun = False
    def __init__(self, *a, **kw):
        _Process.__init__(self, *a, **kw)
        self.directories = {}
        self.doneDirectories = {}

    def addDirectory(self, pathname):
        """I will expect a visit to this directory."""
        # With copious error checking because I don't know how CVS actually
        # works here.
        if self._loginfoBegun:
            log.msg("Unexpected: Directory %s added after I've received"
                    " messages from loginfo." % (pathname,))

        if self.directories.has_key(pathname):
            log.msg("Unexpected: %s is already in my list of directories but"
                    " is being added again." % (pathname,))

        self.directories[pathname] = 'Set'

    def addFiles(self, pathname, files, user, message):
        _Process.addFiles(self, pathname, files, user, message)
        self._loginfoBegun = True
        if self.doneDirectories.has_key(pathname):
            log.msg("Unexpected: directory %s already visited by %s." %
                    (pathname, self.pid))
        elif not self.directories.has_key(pathname):
            # This may be a "cvs add somedirectory" commitinfo doesn't get.
            if (len(files) == 1) and isinstance(files[0], common.Directory):
                log.msg("New directory, assuming process %s isn't doing"
                        " anything else." % (self.pid,))
                self.processComplete()
            elif len(files) == 0:
                log.msg("Zero files, assuming this is a module import and "
                        "that process %s isn't doing anything else."
                        % (self.pid,))
                self.processComplete()
            else:
                # We see this on Imports.
                log.msg("Unexpected: directory %s wasn't in the list." %
                        (pathname,))
                log.msg("(pid %s files %s) %s" % (self.pid, files, message))
        else:
            self.doneDirectories[pathname] = 'Set'
            del self.directories[pathname]
            if not self.directories:
                self.processComplete()

# Process = PollProcProcess
Process = DirectoryListProcess

class PidPlucker(components.Adapter, pb.Root, styles.Versioned):
    """The PB object the loginfo hook delivers its information to.

    I provide the event source for this service.

    I don't think I need authentication because I think I'll be
    listening on unix sockets, and I'm assuming the permissions on
    that socket keep out anyone without CVS access, and if you don't
    trust the people with CVS access...  well, you've got bigger
    problems on your hands than me.
    """

    persistenceVersion = 2

    def remote_loginfo(self, pid, root, pathname, files, user, message):
        IRepositoryTable(self).addProcessEntry(root, pathname,
                                               files, pid, user, message)

    def remote_hasDirectory(self, pid, root, pathname):
        if pathname.startswith(root):
            # XXX: This shouldn't be conditional.  Does this need to be
            # stripped, or doesn't it?
            # +1 for the joining /
            pathname = pathname[len(root)+1:]
        IRepositoryTable(self).addDirectoryForProcess(root, pathname, pid)

    def remote_reconfigure(self, root):
        self.reconfigure(root)

    def reconfigure(self, root=None):
        IRepositoryTable(self).reloadConfig(root)

    def upgradeToVersion2(self):
        # FIXME: Boy howdy is this ever untested!
        self.original = self.service
        del self.service


class LoginfoListener(components.Componentized, MultiService,
                      styles.Versioned):
    persistenceVersion = 3
    name = "loginfo listener"

    def __init__(self):
        components.Componentized.__init__(self)
        MultiService.__init__(self)

    def startService(self):
        r = MultiService.startService(self)
        IRepositoryTable(self).reloadConfig()
        return r

    def upgradeToVersion2(self):
        # self.configFileName = self.configArgs[1]
        del self.configArgs

    def upgradeToVersion3(self):
        # Changed from a twisted.internet.app.ApplicationService
        # to a twisted.application.service.Service!
        del self.plucker
        raise NotImplementedError(
            "Sorry, this class completely changed base classes "
            "in this upgrade.  Conversion not implemented.")

components.registerAdapter(PidPlucker, LoginfoListener, pb.IPBRoot)
components.registerAdapter(RepositoryTable, LoginfoListener, IRepositoryTable)

def updateApplication(ourApp, options):
    """Construct and configure the Fresh daemon.

    @param ourApp: The Application to add my services to.
    @type ourApp: L{twisted.internet.app.Application}
    @param options: command-line options
    @type options: L{Options}
    """

    svc = LoginfoListener()

    if options.opts.get('manhole') is not None:
        # Since the move to makeService, I wonder if using
        # twisted.tap.manhole.makeService isn't a reasonable
        # option now.
        from twisted.cred import portal, checkers
        from twisted.manhole.service import Realm
        from twisted.manhole.service import Service as ManholeService

        p = portal.Portal(
            Realm(ManholeService()),
            [checkers.InMemoryUsernamePasswordDatabaseDontUse(
            manhole=options['manhole'])]
            )
        manservice = strports.service(str(options['manhole-port']),
                                      pb.PBServerFactory(p))
        manservice.setServiceParent(svc)

    svc.setServiceParent(ourApp)

    # TODO: We may operate for many cvs repositories.
    IRepositoryTable(svc).addRepository(Repository(options.opts['root'], svc))

    # TODO: Use t.a.strports
    loginfoServer = internet.UNIXServer(options.opts['pipe'],
                                        pb.PBServerFactory(svc))
    loginfoServer.setServiceParent(svc)
