# -*- Python -*-
# $Id: pidutils.py,v 1.3 2003/02/16 17:24:51 acapnotic Exp $
# by Kevin Turner <acapnotic@foobox.net>
# The contents of this file are public domain.

"""Process-tracking utilities.

Uses /proc filesystem interface, likely Linux-specific.
"""

import errno, os, pwd, stat, string

True = (1==1)
False = not True

class PidException(Exception):
    def __init__(self, pid=None, strerror=None):
        Exception.__init__(self)
        self.pid = pid
        self.strerror = strerror

    def __repr__(self):
        return '<%s(%s) instance at %s>'\
               % (self.__class__, self.pid, id(self))

    def __str__(self):
        return self.strerror or self.__repr__()


class PidStatus:
    """Get status information of a process.

    Provides a mapping interface to the information available in
    /proc/%d%/status.  i.e. to find the name of process 123::

        status = PidStatus(123)
        print status['Name']

    The information is current at the time the PidStatus instance is
    initalized, it is not updated with every access.

    Raises PidException if there doesn't appear to be such a process.

    (Platforms: Linux)
    """
    _d = None

    def __init__(self, pid=None):
        """Get status information for a process.

        If the pid argument is None or not supplied, it defaults to
        the ID of the current process.
        """

        if pid is None:
            pid = os.getpid()

        try:
            f = open('/proc/%d/status' % (pid,))
        except IOError, e:
            if e.errno == errno.ENOENT:
                pe = PidException()
                pe.pid = pid
                pe.errno = errno.ESRCH
                pe.strerror = 'No such PID %d' % (pid,)
                pe.args = (pe.errno, pe.strerror, pe.pid)
                raise pe
            else:
                raise

        self._d = {}

        while 1:
            line = f.readline()
            if not line:
                break
            words = string.split(line)
            key = words[0]
            if key[-1] == ':':
                key = key[:-1]
            if len(words) > 1:
                self._d[key] = string.strip(string.join(words[1:]))
            else:
                self._d[key] = None

    def __getitem__(self, key):

        return self._d[key]


def findParentNamed(name, pid=None, _recursionDepth=0):
    """Finds a process's ancestor with a particular name.

    I walk through a process's ancestors, looking for one whose name
    matches the name argument.  If sucessful, I return the PID of the
    first such process.  If no process with a matching name is found,
    I return None.

    If the process ID from which to start is None or not provided, it
    defaults to the PID of the current process.
    """

    if pid is None:
        pid = os.getpid()

    try:
        s = PidStatus(pid)
    except PidException:
        return None

    if s['Name'] == name:
        return pid
    else:
        pid = int(s['PPid'])
        _recursionDepth = _recursionDepth + 1
        # print 'depth: %3d %5d' % (_recursionDepth, pid)
        return findParentNamed(name, pid, _recursionDepth)


def isDead(pid):
    """Returns True if PId does not exist or is a zombie."""

    try:
        status = PidStatus(pid)
    except PidException, e:
        if e.args[0] == errno.ESRCH:
            # Yes, it's dead, Jim.
            return True
        else:
            # If it's not ENOENT, it's not something we handle.
            raise
    else:
        if string.split(status['State'])[0] == 'Z':
            # Zombie process -- effectively dead.
            return 'Zombie'
        else:
            return False

def owner(pid):
    """Return the login name of the owner of the file."""
    return pwd.getpwuid(os.stat("/proc/%d" % (pid,))[stat.ST_UID])[0]
